# -*- coding: utf-8 -*-
"""
Created on Mon Dec 17 18:17:31 2018

@author: Andreas
"""

import functools

from nion.swift.model import Symbolic


class ComputationUIPanelDelegate(object):

    def __init__(self, api):
        self.__api = api
        self.panel_id = 'ComputationUI-Panel'
        self.panel_name = 'Computation'
        self.panel_positions = ['left', 'right']
        self.panel_position = 'right'
        self.api = api

    def create_panel_widget(self, ui, document_controller):
        self.ui = ui
        self.document_controller = document_controller
        self.__display_item_changed_event_listener = (
                                 document_controller._document_controller.focused_display_item_changed_event.listen(
                                                                                         self.__display_item_changed))
        self.__computation_updated_event_listener = None
        self.__progress_updated_event_listeners = []
        self.column = ui.create_column_widget()

        return self.column

    def __update_progress(self, progress_bar_widget, minimum, maximum, value):
        if minimum != progress_bar_widget.minimum:
            progress_bar_widget.mininum = minimum
        if maximum != progress_bar_widget.maximum:
            progress_bar_widget.maximum = maximum
        progress_bar_widget.value = value

    def __update_computation_ui(self, computations):
        self.column._widget.remove_all()
        for computation in computations:
            create_panel_widget = getattr(computation, 'create_panel_widget', None)
            if create_panel_widget is None:
                compute_class = Symbolic._computation_types.get(computation.processing_id)
                if compute_class:
                    api_computation = self.api._new_api_object(computation)
                    api_computation.api = self.api
                    try:
                        compute_class(api_computation)
                    except IndexError:
                        pass
                    create_panel_widget = getattr(computation, 'create_panel_widget', None)
            if create_panel_widget:
                try:
                    widget = create_panel_widget(self.ui, self.document_controller)
                except Exception as e:
                    print(str(e))
                    import traceback
                    traceback.print_exc()
                else:
                    self.column.add(self.ui.create_label_widget(computation.processing_id))
                    self.column.add_spacing(2)
                    self.column.add(widget)
                    if hasattr(computation, 'progress_updated_event'):
                        progress_bar = self.ui.create_progress_bar_widget()
                        listener = computation.progress_updated_event.listen(functools.partial(self.__update_progress, progress_bar))
                        self.__progress_updated_event_listeners.append(listener)
                        self.column.add_spacing(10)
                        progress_row = self.ui.create_row_widget()
                        progress_row.add_spacing(80)
                        progress_row.add(progress_bar)
                        progress_row.add_spacing(5)
                        self.column.add(progress_row)
                    self.column.add_spacing(15)
        self.column.add_stretch()

    def __get_computations_involved(self, data_item):
        computations = self.document_controller._document_controller.document_model.computations
        computations_involved = []
        for computation in computations:
            for result in computation.results:
                if data_item in result.output_items:
                    if not computation in computations_involved:
                        computations_involved.append(computation)
                    break
            for variable in computation.variables:
                if data_item in variable.input_items:
                    if not computation in computations_involved:
                        computations_involved.append(computation)
                    break
        return computations_involved

    def __display_item_changed(self, display_item):
        data_item = display_item.data_item if display_item else None
        if data_item:
            if self.__computation_updated_event_listener:
                self.__computation_updated_event_listener.close()
                self.__computation_updated_event_listener = None
            for listener in self.__progress_updated_event_listeners:
                listener.close()
            self.__progress_updated_event_listeners = []
            self.__update_computation_ui(self.__get_computations_involved(data_item))

            def computation_updated(computation):
                self.__update_computation_ui(self.__get_computations_involved(data_item))

            self.__computation_updated_event_listener = self.document_controller._document_controller.document_model.computation_updated_event.listen(computation_updated)
        else:
            self.column._widget.remove_all()

    def close(self):
        self.__display_item_changed_event_listener.close()
        self.__display_item_changed_event_listener = None

class ComputationUIExtension(object):
    extension_id = 'nion.extension.computation_ui'

    def __init__(self, api_broker):
        api = api_broker.get_api(version='1', ui_version='1')
        self.__panel_ref = api.create_panel(ComputationUIPanelDelegate(api))

    def close(self):
        self.__panel_ref.close()
        self.__panel_ref = None



import typing
import gettext
import functools
import numpy

from nion.data import Core
from nion.data import DataAndMetadata
from nion.swift.model import Symbolic
from nion.swift import Facade
from nion.typeshed import API_1_0
from nion.utils import Event


_ = gettext.gettext


class AlignMultiSI:
    def __init__(self, computation, **kwargs):
        self.computation = computation
        self.progress_updated_event = Event.Event()

        def create_panel_widget(ui, document_controller):
            def update_align_region_label():
                current_region = self.computation._computation.get_input("align_region")
                haadf_sequence_data_item = self.computation._computation.get_input("haadf_sequence_data_item")
                if current_region and haadf_sequence_data_item:
                    bounds = current_region.bounds
                    shape = haadf_sequence_data_item.xdata.datum_dimension_shape
                    current_region = ((int(bounds[0][0]*shape[0]), int(bounds[0][1]*shape[1])),
                                      (int(bounds[1][0]*shape[0]), int(bounds[1][1]*shape[1])))
                self.align_region_label.text = str(current_region)

            def select_button_clicked():
                for variable in self.computation._computation.variables:
                    if variable.name == "align_region":
                        self.computation._computation.remove_variable(variable)
                graphics = document_controller.target_display.selected_graphics or list()
                align_region = None
                for graphic in graphics:
                    if graphic.graphic_type == "rect-graphic":
                        align_region = graphic
                        break
                if align_region:
                    self.computation._computation.create_input_item("align_region", Symbolic.make_item(align_region._graphic))

            def align_index_finished(text):
                try:
                    index = int(text)
                except ValueError:
                    current_index = self.computation._computation.get_input_value("align_index") or 0
                    self.index_field.text = str(current_index)
                else:
                    self.computation.set_input_value("align_index", index)

            column = ui.create_column_widget()
            row = ui.create_row_widget()

            select_graphics_button = ui.create_push_button_widget("Select align region")
            self.align_region_label = ui.create_label_widget()
            update_align_region_label()
            row.add_spacing(10)
            row.add(select_graphics_button)
            row.add_spacing(5)
            row.add(ui.create_label_widget("Current region: "))
            row.add(self.align_region_label)
            row.add_stretch()
            row.add_spacing(10)

            index_label = ui.create_label_widget("Reference slice index: ")
            current_index = self.computation._computation.get_input_value("align_index") or 0
            self.index_field = ui.create_line_edit_widget(str(current_index))
            self.index_field.on_editing_finished = align_index_finished
            index_row = ui.create_row_widget()
            index_row.add_spacing(10)
            index_row.add(index_label)
            index_row.add(self.index_field)
            index_row.add_spacing(10)
            index_row.add_stretch()

            column.add_spacing(10)
            column.add(row)
            column.add_spacing(5)
            column.add(index_row)

            select_graphics_button.on_clicked = select_button_clicked

            return column

        self.computation._computation.create_panel_widget = create_panel_widget
        self.computation._computation.progress_updated_event = self.progress_updated_event

    def execute(self, si_sequence_data_item: API_1_0.DataItem, haadf_sequence_data_item: API_1_0.DataItem,
                align_index: int, align_region: typing.Optional[API_1_0.Graphic]=None):
        haadf_xdata = haadf_sequence_data_item.xdata
        si_xdata = si_sequence_data_item.xdata
        bounds = None
        if align_region:
            bounds = align_region.bounds
        translations = Core.function_sequence_measure_relative_translation(haadf_xdata,
                                                                           haadf_xdata[align_index],
                                                                           True, bounds=bounds)
        sequence_shape = haadf_sequence_data_item.xdata.sequence_dimension_shape
        data_zeros = (0,) * si_xdata.datum_dimension_count
        c = int(numpy.product(sequence_shape))
        haadf_result_data = numpy.empty_like(haadf_xdata.data)
        si_result_data = numpy.empty_like(si_xdata.data)
        for i in range(c):
            ii = numpy.unravel_index(i, sequence_shape)
            current_xdata = DataAndMetadata.new_data_and_metadata(haadf_xdata.data[ii])
            translation = translations.data[ii]
            haadf_result_data[ii] = Core.function_shift(current_xdata, tuple(translation)).data
            current_xdata = DataAndMetadata.new_data_and_metadata(si_xdata.data[ii])
            si_result_data[ii] = Core.function_shift(current_xdata, tuple(translation) + data_zeros).data
            self.progress_updated_event.fire(0, c, i+1)

        self.__aligned_haadf_sequence = DataAndMetadata.new_data_and_metadata(haadf_result_data,
                                                                              intensity_calibration=haadf_xdata.intensity_calibration,
                                                                              dimensional_calibrations=haadf_xdata.dimensional_calibrations,
                                                                              metadata=haadf_xdata.metadata,
                                                                              data_descriptor=haadf_xdata.data_descriptor)
        self.__aligned_si_sequence = DataAndMetadata.new_data_and_metadata(si_result_data,
                                                                           intensity_calibration=si_xdata.intensity_calibration,
                                                                           dimensional_calibrations=si_xdata.dimensional_calibrations,
                                                                           metadata=si_xdata.metadata,
                                                                           data_descriptor=si_xdata.data_descriptor)

    def commit(self):
        self.computation.set_referenced_xdata("aligned_haadf", self.__aligned_haadf_sequence)
        self.computation.set_referenced_xdata("aligned_si", self.__aligned_si_sequence)


def align_multi_si(api: API_1_0.API, window: API_1_0.DocumentWindow):
    selected_display_items = window._document_controller._get_two_data_sources()
    error_msg = "Select a sequence of spectrum images and a sequence of scanned images in order to use this computation."
    assert selected_display_items[0][0] is not None, error_msg
    assert selected_display_items[1][0] is not None, error_msg
    assert selected_display_items[0][0].data_item is not None, error_msg
    assert selected_display_items[1][0].data_item is not None, error_msg
    assert selected_display_items[0][0].data_item.is_sequence, error_msg
    assert selected_display_items[1][0].data_item.is_sequence, error_msg

    if selected_display_items[0][0].data_item.is_collection:
        si_sequence_data_item = selected_display_items[0][0].data_item
        haadf_sequence_data_item = selected_display_items[1][0].data_item
        align_region = selected_display_items[1][1]
        align_index = selected_display_items[1][0].display_data_channel.sequence_index
    elif selected_display_items[1][0].data_item.is_collection:
        si_sequence_data_item = selected_display_items[1][0].data_item
        haadf_sequence_data_item = selected_display_items[0][0].data_item
        align_region = selected_display_items[0][1]
        align_index = selected_display_items[0][0].display_data_channel.sequence_index
    else:
        raise ValueError(error_msg)

    aligned_haadf = api.library.create_data_item_from_data(numpy.zeros((1,1,1)), title="Aligned {}".format(haadf_sequence_data_item.title))
    aligned_si = api.library.create_data_item_from_data(numpy.zeros((1,1,1)), title="Aligned {}".format(si_sequence_data_item.title))
    inputs = {"si_sequence_data_item": api._new_api_object(si_sequence_data_item),
              "haadf_sequence_data_item": api._new_api_object(haadf_sequence_data_item),
              "align_index": align_index}
    if align_region:
        inputs["align_region"] = api._new_api_object(align_region)
    computation = api.library.create_computation("eels.align_multi_si",
                                                 inputs=inputs,
                                                 outputs={"aligned_haadf": aligned_haadf,
                                                          "aligned_si": aligned_si})
    computation._computation.source = aligned_si._data_item
    window.display_data_item(aligned_haadf)
    window.display_data_item(aligned_si)

Symbolic.register_computation_type("eels.align_multi_si", AlignMultiSI)


class AlignMultiSIMenuItemDelegate:

    def __init__(self, api):
        self.__api = api
        self.menu_id = "eels_menu"  # required, specify menu_id where this item will go
        self.menu_name = _("EELS")  # optional, specify default name if not a standard menu
        self.menu_before_id = "window_menu"  # optional, specify before menu_id if not a standard menu
        self.menu_item_name = _("Align sequence of spectrum images")  # menu item name

    def menu_item_execute(self, window):
        align_multi_si(self.__api, window)


class AlignMultiSIExtension:

    # required for Swift to recognize this as an extension class.
    extension_id = "nion.experimental.align_multi_si"

    def __init__(self, api_broker):
        # grab the api object.
        api = api_broker.get_api(version="~1.0")
        self.__align_multi_si_menu_item_ref = api.create_menu_item(AlignMultiSIMenuItemDelegate(api))

    def close(self):
        # close will be called when the extension is unloaded. in turn, close any references so they get closed. this
        # is not strictly necessary since the references will be deleted naturally when this object is deleted.
        self.__align_multi_si_menu_item_ref.close()
        self.__align_multi_si_menu_item_ref = None



# system imports
import gettext
from nion.swift.model import Symbolic, DataItem
from nion.swift import Facade

# local libraries
from nion.typeshed import API_1_0 as API
from nion.data import xdata_1_0 as xd

import numpy as np
import uuid

from . import ImageChooser
from .DataCache import DataCache

_ = gettext.gettext


class TotalBin4D:
    def __init__(self, computation, **kwargs):
        self.computation = computation

    def execute(self, src):
        try:
            self.__new_xdata = xd.sum(src.xdata, axis=(2, 3))
        except Exception as e:
            import traceback
            traceback.print_exc()
            print(str(e))
            raise

    def commit(self):
        self.computation.set_referenced_xdata('target', self.__new_xdata)


class DarkCorrection4D:
    def __init__(self, computation, **kwargs):
        self.computation = computation
        self.__api = computation.api
        if not hasattr(computation, 'data_cache'):
            computation.data_cache = DataCache()

        def create_panel_widget(ui, document_controller):
            def gain_mode_changed(current_item):
                if current_item != self.computation._computation._get_variable('gain_mode').value:
                    self.computation._computation._get_variable('gain_mode').value = current_item

            def bin_data_changed(check_state):
                if self.computation._computation._get_variable('bin_spectrum').value != (check_state == 'checked'):
                    self.computation._computation._get_variable('bin_spectrum').value = check_state == 'checked'

            def clear_gain_image():
                variable = self.computation._computation._get_variable('gain_image')
                if variable.objects_model.items:
                    variable.objects_model.remove_item(0)

            column = ui.create_column_widget()
            image_chooser, image_changed_listener = ImageChooser.make_image_chooser(document_controller._document_controller,
                                                                                    self.computation._computation,
                                                                                    self.computation._computation._get_variable('gain_image'))
            self.__image_changed_listener = image_changed_listener

            clear_gain_image_button = ui.create_push_button_widget('Clear')

            gain_mode_label = ui.create_label_widget('Gain correction mode: ')
            gain_mode_chooser = ui.create_combo_box_widget()
            gain_mode_chooser.items = ['auto', 'custom', 'off']
            bin_data_checkbox = ui.create_check_box_widget('Bin data to 1D')

            gain_mode_row = ui.create_row_widget()
            gain_mode_row.add(gain_mode_label)
            gain_mode_row.add(gain_mode_chooser)
            gain_mode_row.add_stretch()

            gain_mode_row2 = ui.create_row_widget()
            gain_mode_row2.add_stretch()
            gain_mode_row2._widget.add(image_chooser)
            gain_mode_row2.add_spacing(5)
            gain_mode_row2.add(clear_gain_image_button)

            bin_data_row = ui.create_row_widget()
            bin_data_row.add(bin_data_checkbox)
            bin_data_row.add_stretch()

            column.add(gain_mode_row)
            column.add_spacing(10)
            column.add(gain_mode_row2)
            column.add_spacing(10)
            column.add(bin_data_row)
            column.add_stretch()

            gain_mode_chooser.current_item = self.computation._computation._get_variable('gain_mode').value
            bin_data_checkbox.checked = self.computation._computation._get_variable('bin_spectrum').value

            clear_gain_image_button.on_clicked = clear_gain_image
            gain_mode_chooser.on_current_item_changed = gain_mode_changed
            bin_data_checkbox.on_check_state_changed = bin_data_changed

            return column

        self.computation._computation.create_panel_widget = create_panel_widget

    def execute(self, src1, src2, dark_area_region, crop_region, bin_spectrum, gain_image, gain_mode):
        try:
            xdata = src1.xdata
            metadata = src1.metadata.copy()
            data_shape = np.array(xdata.data.shape)
            dark_area = np.rint(np.array(dark_area_region.bounds) * np.array((data_shape[:2], data_shape[:2]))).astype(np.int)
            crop_area = np.rint(np.array(crop_region.bounds) * np.array((data_shape[2:], data_shape[2:]))).astype(np.int)

            dark_image = xd.sum(xdata[dark_area[0, 0]:dark_area[0, 0]+dark_area[1, 0],
                                      dark_area[0, 1]:dark_area[0, 1]+dark_area[1, 1],
                                      crop_area[0, 0]:crop_area[0, 0]+crop_area[1, 0],
                                      crop_area[0, 1]:crop_area[0, 1]+crop_area[1, 1]], axis=(0, 1))/(dark_area[1,0]*dark_area[1,1])

            self.__new_xdata = xdata[..., crop_area[0, 0]:crop_area[0, 0]+crop_area[1, 0],
                                          crop_area[0, 1]:crop_area[0, 1]+crop_area[1, 1]] - dark_image

            current_gain_image_uuid = metadata.get('hardware_source', {}).get('current_gain_image')
            current_gain_image = None
            if current_gain_image_uuid:
                current_gain_image = self.__api.library.get_data_item_by_uuid(uuid.UUID(current_gain_image_uuid))
            if metadata.get('hardware_source', {}).get('is_gain_corrected'):
                if gain_mode in ('custom', 'off') and current_gain_image:
                    if current_gain_image.xdata.data_shape == xdata.data_shape[2:]:
                        self.__new_xdata /= current_gain_image.xdata[crop_area[0, 0]:crop_area[0, 0]+crop_area[1, 0],
                                                                     crop_area[0, 1]:crop_area[0, 1]+crop_area[1, 1]]

            if ((gain_mode == 'auto' and not metadata.get('hardware_source', {}).get('is_gain_corrected') and current_gain_image) or
                (gain_mode == 'custom' and gain_image)):

                gain_xdata = gain_image[0].xdata if gain_mode == 'custom' else current_gain_image.xdata

                if gain_xdata.data_shape == self.__new_xdata.data_shape[2:]:
                    self.__new_xdata *= gain_xdata
                elif gain_xdata.data_shape == xdata.data_shape[2:]:
                    self.__new_xdata *= gain_xdata[crop_area[0, 0]:crop_area[0, 0]+crop_area[1, 0],
                                                   crop_area[0, 1]:crop_area[0, 1]+crop_area[1, 1]]
                else:
                    raise ValueError('Shape of gain image has to match last two dimensions of input data.')
                del gain_xdata

            if bin_spectrum:
                self.__new_xdata = xd.sum(self.__new_xdata, axis=2)

            metadata['nion.dark_correction_4d.parameters'] = {'src1': src1._data_item.write_to_dict(),
                                                              'src2': src2._data_item.write_to_dict(),
                                                              'dark_area_region': dark_area_region._graphic.write_to_dict(),
                                                              'crop_region': crop_region._graphic.write_to_dict(),
                                                              'bottom_dark_region_region': crop_region._graphic.write_to_dict(),
                                                              'bin_spectrum': bin_spectrum,
                                                              'gain_image': gain_image[0].data_item.write_to_dict() if gain_image else None,
                                                              'gain_mode': gain_mode}
            self.__new_xdata.metadata.update(metadata)
        except Exception as e:
            print(str(e))
            import traceback
            traceback.print_exc()

    def commit(self):
        self.computation.set_referenced_xdata('target', self.__new_xdata)

class DarkCorrection4DMenuItem:

    menu_id = "4d_tools_menu"  # required, specify menu_id where this item will go
    menu_name = _("4D Tools") # optional, specify default name if not a standard menu
    menu_before_id = "window_menu" # optional, specify before menu_id if not a standard menu
    menu_item_name = _("4D Dark Correction")  # menu item name

    #DocumentModel.DocumentModel.register_processing_descriptions(correct_dark_processing_descriptions)
    #DocumentModel.DocumentModel.register_processing_descriptions(calculate_average_processing_descriptions)
    def __init__(self, api):
        self.__api = api
        self.__computation_data_items = dict()
        self.__tool_tip_boxes = list()

    def __display_item_changed(self, display_item):
        data_item = display_item.data_item if display_item else None
        if data_item:
            tip_id = self.__computation_data_items.get(data_item)
            if tip_id:
                self.__show_tool_tips(tip_id)

    def __show_tool_tips(self, tip_id='source', timeout=30):
        for box in self.__tool_tip_boxes:
            box.remove_now()
        self.__tool_tip_boxes = list()
        if tip_id == 'source':
            text = 'Use the "Crop" graphic to crop the camera images.\nSelect the "Total bin..." or "4D dark correction..." data item for further options.'
        elif tip_id == 'total bin':
            text = 'Use the "Dark subtract area" graphic to select the are that was not illuminated. The sum of all camera frames will be the dark image.\nSelect the source data item or the "4D dark correction..." data item for further options.'
        elif tip_id == 'corrected':
            text = 'In the "Computation" panel (Window -> Computation) you find the settings. Custom gain correction mode uses the gain image from the drag-and-drop area.\n"Bin data to 1D" will ouput 1D spectra instead of 2D camera frames in the result data item.\nSelect the source data item or the "Total bin..." data item for further options.'
        elif tip_id == 'wrong_shape':
            text = 'This computation only works for 4D-data.'
        else:
            return
        document_controller = self.__api.application.document_windows[0]
        #box = document_controller.show_tool_tip_box(text, timeout)
        workspace = document_controller._document_controller.workspace_controller
        box = workspace.pose_tool_tip_box(text, timeout)
        self.__tool_tip_boxes.append(box)

    def menu_item_execute(self, window: API.DocumentWindow) -> None:
        document_controller = window._document_controller
        selected_display_item = document_controller.selected_display_item
        data_item = (selected_display_item.data_items[0] if
                     selected_display_item and len(selected_display_item.data_items) > 0 else None)

        if data_item:
            api_data_item = Facade.DataItem(data_item)
            if not api_data_item.xdata.is_data_4d:
                self.__show_tool_tips('wrong_shape')
                return
            total_bin_data_item = self.__api.library.create_data_item(title='Total bin 4D of ' + data_item.title)
            computation = self.__api.library.create_computation('nion.total_bin_4d_SI',
                                                                inputs={'src': api_data_item},
                                                                outputs={'target': total_bin_data_item})
            computation._computation.source = total_bin_data_item._data_item
            #computation._computation.mark_update()

            total_bin_display_item = document_controller.document_model.get_display_item_for_data_item(
                                                                                                  total_bin_data_item._data_item)
            document_controller.show_display_item(total_bin_display_item)
            dark_subtract_area_graphic = total_bin_data_item.add_rectangle_region(0.8, 0.5, 0.4, 1.0)
            dark_subtract_area_graphic.label = 'Dark subtract area'
            crop_region = api_data_item.add_rectangle_region(0.5, 0.5, 1.0, 1.0)
            crop_region.label = 'Crop'

            dark_subtract_area_graphic._graphic.is_bounds_constrained = True
            crop_region._graphic.is_bounds_constrained = True
            dark_corrected_data_item = Facade.DataItem(DataItem.DataItem(large_format=True))
            self.__api.library._document_model.append_data_item(dark_corrected_data_item._data_item)
            dark_corrected_data_item._data_item.session_id = self.__api.library._document_model.session_id
            dark_corrected_data_item.title = '4D dark correction of ' + data_item.title
            computation = self.__api.library.create_computation('nion.dark_correction_4d',
                                                                inputs={'src1': api_data_item,
                                                                        'src2': total_bin_data_item,
                                                                        'dark_area_region': dark_subtract_area_graphic,
                                                                        'crop_region': crop_region,
                                                                        'bin_spectrum': True,
                                                                        'gain_image': [],
                                                                        'gain_mode': 'custom'},
                                                                outputs={'target': dark_corrected_data_item})
            computation._computation.source = dark_corrected_data_item._data_item
            dark_corrected_display_item = document_controller.document_model.get_display_item_for_data_item(
                                                                                             dark_corrected_data_item._data_item)
            document_controller.show_display_item(dark_corrected_display_item)
            self.__computation_data_items.update({data_item: 'source',
                                                  total_bin_data_item._data_item: 'total bin',
                                                  dark_corrected_data_item._data_item: 'corrected'})
            self.__show_tool_tips()
            self.__display_item_changed_event_listener = (
                           document_controller.focused_display_item_changed_event.listen(self.__display_item_changed))

class DarkCorrection4DExtension:

    # required for Swift to recognize this as an extension class.
    extension_id = "nion.extension.4d_dark_correction"

    def __init__(self, api_broker):
        # grab the api object.
        api = api_broker.get_api(version="1", ui_version="1")
        # be sure to keep a reference or it will be closed immediately.
        self.__menu_item_ref = api.create_menu_item(DarkCorrection4DMenuItem(api))

    def close(self):
        self.__menu_item_ref.close()

Symbolic.register_computation_type('nion.total_bin_4d_SI', TotalBin4D)
Symbolic.register_computation_type('nion.dark_correction_4d', DarkCorrection4D)



import threading
import time
import numpy as np

class DataCache:
    def __init__(self, lifetime=30, modify_data_fn=None):
        self.__cached_data = None
        self.__cached_uuid = None
        self.__last_requested = time.time()

        self.lifetime = lifetime
        self.modify_data_fn = modify_data_fn if callable(modify_data_fn) else np.array

        self.__thread = threading.Thread(target=self.__cache_loop, daemon=True)
        self.__lock = threading.Lock()
        self.__thread.start()

    def __cache_loop(self):
        while True:
            with self.__lock:
                now = time.time()
                if now - self.__last_requested > self.lifetime:
                    self.__cached_data = None
                    self.__cached_uuid = None
            time.sleep(0.5)

    def get_cached_data(self, data_source):
        uuid = str(data_source.uuid)
        with self.__lock:
            if self.__cached_uuid != uuid:
                xdata = data_source.xdata
                self.__cached_uuid = uuid
                self.__cached_data = self.modify_data_fn(xdata.data)
                #self.__cached_data = np.reshape(xdata.data, xdata.data.shape[:2] + (-1,))
            self.__last_requested = time.time()
            return self.__cached_data



# system imports
import gettext
import logging
import copy

import numpy

# local libraries
from nion.typeshed import API_1_0 as API
from nion.data import xdata_1_0 as xd
from nion.swift import Facade
from nion.swift.model import Symbolic
from nion.data import DataAndMetadata
from nion.data import Core

_ = gettext.gettext


class AffineTransformImage:
    label = _("Affine Transform Image")
    inputs = {
        "src_data_item": {"label": _("Source")},
        "vector_a": {"label": _("Vector a")},
        "vector_b": {"label": _("Vector b")},
        }
    outputs = {
        "target": {"label": _("Affine Transformed")},
    }
    def __init__(self, computation, **kwargs):
        self.computation = computation

    def execute(self, src_data_item, vector_a, vector_b, **kwargs) -> None:
        try:
            vector_a = vector_a[0]
            vector_b = vector_b[0]
            vector_1 = (numpy.array(vector_a.end) - numpy.array(vector_a.start)) * 4.0
            vector_2 = (numpy.array(vector_b.end) - numpy.array(vector_b.start)) * 4.0

            matrix = numpy.array(((vector_1[1], vector_2[1]),
                                  (vector_1[0], vector_2[0])))
            xdata = src_data_item.xdata
            if not xdata:
                return
            if (xdata.is_sequence or xdata.is_collection) and xdata.collection_dimension_count != 2:
                navigation_dimensions = 0
                if xdata.is_sequence:
                    navigation_dimensions += 1
                navigation_dimensions += xdata.collection_dimension_count
                navigation_dimensions_shape = xdata.data.shape[:navigation_dimensions]
                result_data = numpy.zeros_like(xdata.data)
                new_coords = Core.calculate_coordinates_for_affine_transform(xdata.data[numpy.unravel_index(0, navigation_dimensions_shape)], matrix)
                for i in range(numpy.prod(navigation_dimensions_shape)):
                    index = numpy.unravel_index(i, navigation_dimensions_shape)
                    transformed = xd.warp(xdata.data[index], new_coords)
                    result_data[index] = transformed.data
                self._affine_transformed_xdata = DataAndMetadata.new_data_and_metadata(result_data,
                                                                                        intensity_calibration=xdata.intensity_calibration,
                                                                                        dimensional_calibrations=xdata.dimensional_calibrations,
                                                                                        data_descriptor=xdata.data_descriptor)
            elif xdata.collection_dimension_count == 2: # Assume we want to distort allong the collection dimension in this case
                navigation_dimensions_shape = tuple()
                if xdata.is_sequence:
                    navigation_dimensions_shape = (xdata.data_shape[0],) + xdata.data_shape[3:]
                else:
                    navigation_dimensions_shape = xdata.data_shape[2:]

                if xdata.is_data_rgb_type:
                    navigation_dimensions_shape = navigation_dimensions_shape[:-1]
                index = numpy.unravel_index(0, navigation_dimensions_shape)
                index = ((index[0],) if xdata.is_sequence else tuple()) + (slice(None), slice(None)) + index[-xdata.datum_dimension_count:]
                new_coords = Core.calculate_coordinates_for_affine_transform(xdata.data[index], matrix)
                result_data = numpy.zeros_like(xdata.data)
                for i in range(numpy.prod(navigation_dimensions_shape)):
                    index = numpy.unravel_index(i, navigation_dimensions_shape)
                    index = ((index[0],) if xdata.is_sequence else tuple()) + (slice(None), slice(None)) + index[-xdata.datum_dimension_count:]
                    transformed = xd.warp(xdata.data[index], new_coords)
                    result_data[index] = transformed.data
                self._affine_transformed_xdata = DataAndMetadata.new_data_and_metadata(result_data,
                                                                                        intensity_calibration=xdata.intensity_calibration,
                                                                                        dimensional_calibrations=xdata.dimensional_calibrations,
                                                                                        data_descriptor=xdata.data_descriptor)
            else:
                self._affine_transformed_xdata = xd.affine_transform(xdata, matrix)
            metadata = self._affine_transformed_xdata.metadata
            metadata["nion.affine_transform_image.transformation_matrix"] = matrix.tolist()
            self._affine_transformed_xdata._set_metadata(metadata)
        except:
            import traceback
            traceback.print_exc()
            raise

    def commit(self):
        self.computation.set_referenced_xdata("target", self._affine_transformed_xdata)


class AffineTransformMenuItem:
    menu_id = "_processing_menu"
    menu_item_name = _("Affine Transform image")

    def __init__(self, api):
        self.__api = api

    def menu_item_execute(self, window: API.DocumentWindow) -> None:
        data_item = window.target_data_item

        if not data_item:
            return

        vector_a = data_item.add_line_region(0.5, 0.5, 0.5, 0.75)
        vector_a.label = "Vector a"
        vector_a._graphic.end_arrow_enabled = True
        vector_b = data_item.add_line_region(0.5, 0.5, 0.75, 0.5)
        vector_b.label = "Vector b"
        vector_b._graphic.end_arrow_enabled = True

        result_data_item = self.__api.library.create_data_item_from_data_and_metadata(DataAndMetadata.new_data_and_metadata(numpy.zeros_like(data_item.data), data_descriptor=data_item.xdata.data_descriptor), title="Affine Transformed {}".format(data_item.title))
        self.__api.library.create_computation("nion.affine_transform_image",
                                              inputs={"src_data_item": data_item, "vector_a": [vector_a], "vector_b": [vector_b]},
                                              outputs={"target": result_data_item})
        window.display_data_item(result_data_item)

class CopyAffineTransformMenuItem:
    menu_id = "_processing_menu"
    menu_item_name = _("Copy Affine Transformation")

    def __init__(self, api):
        self.__api = api

    def menu_item_execute(self, window: API.DocumentWindow) -> None:
        document_controller = window._document_controller
        selected_display_items = document_controller.selected_display_items
        data_items = list()
        src_vector_a = None
        src_vector_b = None
        for i, display_item in enumerate(selected_display_items):
            data_item = display_item.data_items[0] if display_item and len(display_item.data_items) > 0 else None
            if data_item and src_vector_a is None:
                for computation in self.__api.library._document_model.computations:
                    if computation.processing_id == "nion.affine_transform_image":
                        if computation.get_input("src_data_item") == data_item:
                            src_vector_a = computation.get_input("vector_a")
                            if src_vector_a is not None:
                                src_vector_a = self.__api._new_api_object(src_vector_a[0])
                            src_vector_b = computation.get_input("vector_b")
                            if src_vector_b is not None:
                                src_vector_b = self.__api._new_api_object(src_vector_b[0])
                            break
                else:
                    data_items.append(self.__api._new_api_object(data_item))

            elif data_item:
                data_items.append(self.__api._new_api_object(data_item))
        if src_vector_a is None:
            return

        for data_item in data_items:
            vector_a = data_item.add_line_region(src_vector_a.start[0], src_vector_a.start[1], src_vector_a.end[0], src_vector_a.end[1])
            vector_a.label = "Vector a"
            vector_a._graphic.end_arrow_enabled = True
            vector_b = data_item.add_line_region(src_vector_b.start[0], src_vector_b.start[1], src_vector_b.end[0], src_vector_b.end[1])
            vector_b.label = "Vector b"
            vector_b._graphic.end_arrow_enabled = True

            result_data_item = self.__api.library.create_data_item_from_data_and_metadata(DataAndMetadata.new_data_and_metadata(numpy.zeros_like(data_item.data), data_descriptor=data_item.xdata.data_descriptor), title="Affine Transformed {}".format(data_item.title))
            self.__api.library.create_computation("nion.affine_transform_image",
                                                  inputs={"src_data_item": data_item, "vector_a": [vector_a], "vector_b": [vector_b]},
                                                  outputs={"target": result_data_item})
            window.display_data_item(result_data_item)


class AffineTransformExtension:
    extension_id = "nion.extension.affine_transform"

    def __init__(self, api_broker):
        api = api_broker.get_api(version="1", ui_version="1")

        self.__affine_transform_menu_item_ref = api.create_menu_item(AffineTransformMenuItem(api))
        self.__copy_affine_transform_menu_item_ref = api.create_menu_item(CopyAffineTransformMenuItem(api))

    def close(self):
        self.__affine_transform_menu_item_ref.close()
        self.__copy_affine_transform_menu_item_ref.close()


Symbolic.register_computation_type("nion.affine_transform_image", AffineTransformImage)


# system imports
import gettext
from nion.swift.model import DataItem, Symbolic
from nion.swift import Facade

# local libraries
from nion.typeshed import API_1_0 as API
from nion.data import xdata_1_0 as xd

from . import ImageChooser
from .DataCache import DataCache

import numpy as np
import uuid

_ = gettext.gettext


class CalculateAverage4D:
    def __init__(self, computation, **kwargs):
        self.computation = computation

    def execute(self, src):
        self.__new_xdata = xd.sum(src.xdata, axis=(0, 1))

    def commit(self):
        self.computation.set_referenced_xdata('target', self.__new_xdata)


class FramewiseDarkCorrection:
    def __init__(self, computation, **kwargs):
        self.computation = computation
        self.__api = computation.api
        if not hasattr(computation, 'data_cache'):
            computation.data_cache = DataCache()

        def create_panel_widget(ui, document_controller):
            def gain_mode_changed(current_item):
                if current_item != self.computation._computation._get_variable('gain_mode').value:
                    self.computation._computation._get_variable('gain_mode').value = current_item

            def bin_data_changed(check_state):
                if self.computation._computation._get_variable('bin_spectrum').value != (check_state == 'checked'):
                    self.computation._computation._get_variable('bin_spectrum').value = check_state == 'checked'

            def clear_gain_image():
                variable = self.computation._computation._get_variable('gain_image')
                if variable.objects_model.items:
                    variable.objects_model.remove_item(0)

            column = ui.create_column_widget()
            image_chooser, image_changed_listener = ImageChooser.make_image_chooser(document_controller._document_controller,
                                                                                    self.computation._computation,
                                                                                    self.computation._computation._get_variable('gain_image'))
            self.__image_changed_listener = image_changed_listener

            clear_gain_image_button = ui.create_push_button_widget('Clear')

            gain_mode_label = ui.create_label_widget('Gain correction mode: ')
            gain_mode_chooser = ui.create_combo_box_widget()
            gain_mode_chooser.items = ['auto', 'custom', 'off']
            bin_data_checkbox = ui.create_check_box_widget('Bin data to 1D')

            gain_mode_row = ui.create_row_widget()
            gain_mode_row.add(gain_mode_label)
            gain_mode_row.add(gain_mode_chooser)
            gain_mode_row.add_stretch()

            gain_mode_row2 = ui.create_row_widget()
            gain_mode_row2.add_stretch()
            gain_mode_row2._widget.add(image_chooser)
            gain_mode_row2.add_spacing(5)
            gain_mode_row2.add(clear_gain_image_button)

            bin_data_row = ui.create_row_widget()
            bin_data_row.add(bin_data_checkbox)
            bin_data_row.add_stretch()

            column.add(gain_mode_row)
            column.add_spacing(10)
            column.add(gain_mode_row2)
            column.add_spacing(10)
            column.add(bin_data_row)
            column.add_stretch()

            gain_mode_chooser.current_item = self.computation._computation._get_variable('gain_mode').value
            bin_data_checkbox.checked = self.computation._computation._get_variable('bin_spectrum').value

            clear_gain_image_button.on_clicked = clear_gain_image
            gain_mode_chooser.on_current_item_changed = gain_mode_changed
            bin_data_checkbox.on_check_state_changed = bin_data_changed

            return column

        self.computation._computation.create_panel_widget = create_panel_widget

    def execute(self, src1, src2, spectrum_region, top_dark_region, bottom_dark_region, bin_spectrum,
                gain_image, gain_mode):
        try:
            data = src1.xdata.data
            data_shape = np.array(data.shape)
            metadata = src1.metadata.copy()
            sensor_readout_area = metadata.get('hardware_source', {}).get('sensor_readout_area')
            sensor_dimensions = metadata.get('hardware_source', {}).get('sensor_dimensions')
            if sensor_readout_area and sensor_dimensions:
                cam_center = sensor_dimensions['height']/2 - sensor_readout_area['top']
            else:
                cam_center = int(round(data_shape[-2]/2))
            spectrum_area = np.rint(np.array(spectrum_region.bounds) * data_shape[2:]).astype(np.int)
            top_dark_area = np.rint(np.array(top_dark_region.bounds) * data_shape[2:]).astype(np.int)
            bottom_dark_area = np.rint(np.array(bottom_dark_region.bounds) * data_shape[2:]).astype(np.int)
            spectrum_range_y = np.array((spectrum_area[0,0], spectrum_area[0,0] + spectrum_area[1,0]))
            spectrum_range_x = np.array((spectrum_area[0,1], spectrum_area[0,1] + spectrum_area[1,1]))
            top_dark_area_range_y = np.array((top_dark_area[0,0], top_dark_area[0,0] + top_dark_area[1, 0]))
            bottom_dark_area_range_y = np.array((bottom_dark_area[0,0], bottom_dark_area[0,0] + bottom_dark_area[1, 0]))

            # undo gain correction if neccessary
            current_gain_image_uuid = metadata.get('hardware_source', {}).get('current_gain_image')
            current_gain_image = None
            if current_gain_image_uuid:
                current_gain_image = self.__api.library.get_data_item_by_uuid(uuid.UUID(current_gain_image_uuid))
            if metadata.get('hardware_source', {}).get('is_gain_corrected') and current_gain_image:
                if current_gain_image.xdata.data_shape == src1.xdata.data_shape[2:]:
                    data = data/current_gain_image.xdata.data

            if (cam_center >= spectrum_range_y).all(): # spectrum is above center
                dark_image = np.mean(data[..., top_dark_area_range_y[0]:top_dark_area_range_y[1],
                                          spectrum_range_x[0]:spectrum_range_x[1]], axis=-2, keepdims=True)
                corrected_image = (data[..., spectrum_range_y[0]:spectrum_range_y[1], spectrum_range_x[0]:spectrum_range_x[1]] -
                                   np.repeat(dark_image, spectrum_range_y[1]-spectrum_range_y[0], axis=-2))
            elif (cam_center <= spectrum_range_y).all(): # spectrum is below center
                dark_image = np.mean(data[..., bottom_dark_area_range_y[0]:bottom_dark_area_range_y[1],
                                          spectrum_range_x[0]:spectrum_range_x[1]], axis=-2, keepdims=True)
                corrected_image = (data[..., spectrum_range_y[0]:spectrum_range_y[1], spectrum_range_x[0]:spectrum_range_x[1]] -
                                   np.repeat(dark_image, spectrum_range_y[1]-spectrum_range_y[0], axis=-2))
            else: # spectrum is on top of center
                dark_image = np.mean(data[..., top_dark_area_range_y[0]:top_dark_area_range_y[1],
                                          spectrum_range_x[0]:spectrum_range_x[1]], axis=-2, keepdims=True)
                corrected_image_top = (data[..., spectrum_range_y[0]:cam_center, spectrum_range_x[0]:spectrum_range_x[1]] -
                                       np.repeat(dark_image, cam_center-spectrum_range_y[0], axis=-2))
                dark_image = np.mean(data[..., bottom_dark_area_range_y[0]:bottom_dark_area_range_y[1],
                                          spectrum_range_x[0]:spectrum_range_x[1]], axis=-2, keepdims=True)
                corrected_image_bot = (data[..., cam_center:spectrum_range_y[1], spectrum_range_x[0]:spectrum_range_x[1]] -
                                       np.repeat(dark_image, spectrum_range_y[1]-cam_center, axis=-2))
                corrected_image = np.concatenate((corrected_image_top, corrected_image_bot), axis=-2)
                del corrected_image_top
                del corrected_image_bot
            del data
            del dark_image # don't hold references to unused objects so that garbage collector can free the memory

            if ((gain_mode == 'auto' and current_gain_image) or # apply gain correction if needed
                (gain_mode == 'custom' and gain_image)):

                gain_xdata = gain_image[0].xdata if gain_mode == 'custom' else current_gain_image.xdata

                if gain_xdata.data_shape == corrected_image.shape[2:]:
                    corrected_image *= gain_xdata.data
                elif gain_xdata.data_shape == src1.xdata.data_shape[2:]:
                    corrected_image *= gain_xdata.data[spectrum_range_y[0]:spectrum_range_y[1],
                                                       spectrum_range_x[0]:spectrum_range_x[1]]
                else:
                    raise ValueError('Shape of gain image has to match last two dimensions of input data.')
                del gain_xdata

            dimensional_calibrations = src1.xdata.dimensional_calibrations.copy()
            if bin_spectrum:
                corrected_image = np.sum(corrected_image, axis=-2)
                dimensional_calibrations = dimensional_calibrations[:2] + dimensional_calibrations[3:]

            data_descriptor = self.__api.create_data_descriptor(False, 2, 1 if bin_spectrum else 2)
            metadata['nion.framewise_dark_correction.parameters'] = {'src1': src1._data_item.write_to_dict(),
                                                                     'src2': src2._data_item.write_to_dict(),
                                                                     'spectrum_region': spectrum_region._graphic.write_to_dict(),
                                                                     'top_dark_region': top_dark_region._graphic.write_to_dict(),
                                                                     'bottom_dark_region_region': bottom_dark_region._graphic.write_to_dict(),
                                                                     'bin_spectrum': bin_spectrum,
                                                                     'gain_image': gain_image[0].data_item.write_to_dict() if gain_image else None,
                                                                     'gain_mode': gain_mode}

            self.__new_xdata = self.__api.create_data_and_metadata(corrected_image,
                                                                   intensity_calibration=src1.xdata.intensity_calibration,
                                                                   dimensional_calibrations=dimensional_calibrations,
                                                                   data_descriptor=data_descriptor,
                                                                   metadata=metadata)
        except Exception as e:
            print(str(e))
            import traceback
            traceback.print_exc()

    def commit(self):
        self.computation.set_referenced_xdata('target', self.__new_xdata)

class FramewiseDarkMenuItem:

    menu_id = "4d_tools_menu"  # required, specify menu_id where this item will go
    menu_name = _("4D Tools") # optional, specify default name if not a standard menu
    menu_before_id = "window_menu" # optional, specify before menu_id if not a standard menu
    menu_item_name = _("Framewise Dark Correction")  # menu item name

    #DocumentModel.DocumentModel.register_processing_descriptions(correct_dark_processing_descriptions)
    #DocumentModel.DocumentModel.register_processing_descriptions(calculate_average_processing_descriptions)
    def __init__(self, api):
        self.__api = api
        self.__computation_data_items = dict()
        self.__tool_tip_boxes = list()

    def __display_item_changed(self, display_item):
        data_item = display_item.data_item if display_item else None
        if data_item:
            tip_id = self.__computation_data_items.get(data_item)
            if tip_id:
                self.__show_tool_tips(tip_id)

    def __show_tool_tips(self, tip_id='source', timeout=30):
        for box in self.__tool_tip_boxes:
            box.remove_now()
        self.__tool_tip_boxes = list()
        if tip_id == 'source':
            text = 'In the "Computation" panel (Window -> Computation) you find the settings. Custom gain correction mode uses the gain image from the drag-and-drop area.\n"Bin data to 1D" will ouput 1D spectra instead of 2D camera frames in the result data item.\nSelect the "Frame average..." data item for further options.'
        elif tip_id == 'average':
            text = 'Use the three graphics to select the area of the detector that contains the spectra and a part of the detector that was not illuminated on top and below the spectrum (top dark area and bottom dark area).'
        elif tip_id == 'corrected':
            text = 'In the "Computation" panel (Window -> Computation) you find the settings. Custom gain correction mode uses the gain image from the drag-and-drop area.\n"Bin data to 1D" will ouput 1D spectra instead of 2D camera frames in the result data item.\nSelect the "Frame average..." data item for further options.'
        elif tip_id == 'wrong_shape':
            text = 'This computation only works for 4D-data.'
        else:
            return
        document_controller = self.__api.application.document_windows[0]
        #box = document_controller.show_tool_tip_box(text, timeout)
        workspace = document_controller._document_controller.workspace_controller
        box = workspace.pose_tool_tip_box(text, timeout)
        self.__tool_tip_boxes.append(box)

    def menu_item_execute(self, window: API.DocumentWindow) -> None:
        document_controller = window._document_controller
        selected_display_item = document_controller.selected_display_item
        data_item = (selected_display_item.data_items[0] if selected_display_item and
                     len(selected_display_item.data_items) > 0 else None)

        if data_item:
            api_data_item = Facade.DataItem(data_item)
            if not api_data_item.xdata.is_data_4d:
                self.__show_tool_tips('wrong_shape')
                return
            average_data_item = self.__api.library.create_data_item(title='Frame average of ' + data_item.title)
            computation = self.__api.library.create_computation('nion.calculate_4d_average',
                                                                inputs={'src': api_data_item},
                                                                outputs={'target': average_data_item})
            computation._computation.source = average_data_item._data_item
            average_display_item = document_controller.document_model.get_display_item_for_data_item(average_data_item._data_item)
            document_controller.show_display_item(average_display_item)
            spectrum_graphic = average_data_item.add_rectangle_region(0.5, 0.5, 0.1, 1.0)
            spectrum_graphic.label = 'Spectrum'
            bottom_dark_graphic = average_data_item.add_rectangle_region(0.7, 0.5, 0.1, 1.0)
            bottom_dark_graphic.label = 'Bottom dark area'
            top_dark_graphic = average_data_item.add_rectangle_region(0.3, 0.5, 0.1, 1.0)
            top_dark_graphic.label = 'Top dark area'
            spectrum_graphic._graphic.is_bounds_constrained = True
            bottom_dark_graphic._graphic.is_bounds_constrained = True
            top_dark_graphic._graphic.is_bounds_constrained = True

            dark_corrected_data_item = Facade.DataItem(DataItem.DataItem(large_format=True))
            self.__api.library._document_model.append_data_item(dark_corrected_data_item._data_item)
            dark_corrected_data_item._data_item.session_id = self.__api.library._document_model.session_id
            dark_corrected_data_item.title = 'Framewise dark correction of ' + data_item.title
            computation = self.__api.library.create_computation('nion.framewise_dark_correction',
                                                                inputs={'src1': api_data_item,
                                                                        'src2': average_data_item,
                                                                        'spectrum_region': spectrum_graphic,
                                                                        'top_dark_region': top_dark_graphic,
                                                                        'bottom_dark_region': bottom_dark_graphic,
                                                                        'bin_spectrum': True,
                                                                        'gain_image': [],
                                                                        'gain_mode': 'custom'},
                                                                outputs={'target': dark_corrected_data_item})
            computation._computation.source = dark_corrected_data_item._data_item
            dark_corrected_display_item = document_controller.document_model.get_display_item_for_data_item(
                                                                                              dark_corrected_data_item._data_item)
            document_controller.show_display_item(dark_corrected_display_item)
            self.__computation_data_items.update({data_item: 'source',
                                                  average_data_item._data_item: 'average',
                                                  dark_corrected_data_item._data_item: 'corrected'})
            self.__show_tool_tips()
            self.__display_item_changed_event_listener = (
                           document_controller.focused_display_item_changed_event.listen(self.__display_item_changed))

class FramewiseDarkExtension:

    # required for Swift to recognize this as an extension class.
    extension_id = "nion.extension.framewise_dark_correction"

    def __init__(self, api_broker):
        # grab the api object.
        api = api_broker.get_api(version="1", ui_version="1")
        # be sure to keep a reference or it will be closed immediately.
        self.__menu_item_ref = api.create_menu_item(FramewiseDarkMenuItem(api))

    def close(self):
        self.__menu_item_ref.close()

Symbolic.register_computation_type('nion.calculate_4d_average', CalculateAverage4D)
Symbolic.register_computation_type('nion.framewise_dark_correction', FramewiseDarkCorrection)



import gettext
import logging

import numpy
import scipy.optimize

from nion.data import DataAndMetadata
from nion.swift.model import Symbolic
from nion.typeshed import API_1_0 as API
from nion.swift import Facade

_ = gettext.gettext


class MakeIDPC:
    label = _("Make iDPC from DPC")
    inputs = {"src": {"label": _("DPC Data Item")},
              "gradient_x_index": {"label": _("DPC x-slice")},
              "gradient_y_index": {"label": _("DPC y-slice")},
              "rotation": {"label": _("Rotation (deg) or 'None' for automatic")},
              "crop_region": {"label": _("Crop")},
              }
    outputs = {"output": {"label": _("iDPC")}}

    def __init__(self, computation, **kwargs):
        self.computation = computation
        self.__result_xdata = None

    def __calculate_curl(self, rotation, com_x, com_y):
        com_x_rotated = com_x * numpy.cos(rotation) - com_y * numpy.sin(rotation)
        com_y_rotated = com_x * numpy.sin(rotation) + com_y * numpy.cos(rotation)
        curl_com = numpy.gradient(com_y_rotated, axis=1) - numpy.gradient(com_x_rotated, axis=0)
        return numpy.mean(curl_com**2)

    def execute(self, src, gradient_x_index, gradient_y_index, rotation, crop_region, **kwargs) -> None:
        try:
            dpc_xdata = src.xdata
            assert dpc_xdata.is_datum_2d
            assert dpc_xdata.is_sequence or dpc_xdata.collection_dimension_count == 1
            gradx = dpc_xdata.data[gradient_x_index]
            grady = dpc_xdata.data[gradient_y_index]
            top_x = crop_region.bounds[0][1] * gradx.shape[1]
            top_y = crop_region.bounds[0][0] * gradx.shape[0]
            crop_slices = (slice(int(top_y), int(top_y + crop_region.bounds[1][0] * gradx.shape[0])),
                           slice(int(top_x), int(top_x + crop_region.bounds[1][1] * gradx.shape[1])))
            # Subtract the mean of each component so that we remove any global offset
            gradx = gradx[crop_slices] - numpy.mean(gradx[crop_slices])
            grady = grady[crop_slices] - numpy.mean(grady[crop_slices])
            # Don't use "if rotation" here because that would also calculate rotation for a given value of 0
            if rotation is None or rotation == "None" or rotation == "":
                res = scipy.optimize.minimize_scalar(self.__calculate_curl, 0, args=(gradx, grady), bounds=(0, numpy.pi*2), method='bounded')
                if res.success:
                    rotation = res.x
                    logging.info(f'Calculated optimal roation: {rotation*180/numpy.pi:.1f} degree.')
                else:
                    logging.warning(f'Could not find the optimal rotation. Optimize error: {res.message}\nUsing rotation=0 as default.')
                    rotation = 0
            else:
                rotation = float(rotation) / 180.0 * numpy.pi

            gradx_rotated = gradx * numpy.cos(rotation) - grady * numpy.sin(rotation)
            grady_rotated = gradx * numpy.sin(rotation) + grady * numpy.cos(rotation)
            freq_v = numpy.fft.fftfreq(gradx.shape[-2], d=dpc_xdata.dimensional_calibrations[-2].scale)
            freq_u = numpy.fft.fftfreq(gradx.shape[-1], d=dpc_xdata.dimensional_calibrations[-1].scale)
            freqs = numpy.meshgrid(freq_u, freq_v)

            fft_idpc = (numpy.fft.fft2(gradx_rotated) * freqs[0] + numpy.fft.fft2(grady_rotated) * freqs[1]) / (1j * (freqs[0]**2 + freqs[1]**2))
            fft_idpc[numpy.isnan(fft_idpc)] = 0
            self.__result_xdata = DataAndMetadata.new_data_and_metadata(numpy.real(numpy.fft.ifft2(fft_idpc)),
                                                                        intensity_calibration=dpc_xdata.intensity_calibration,
                                                                        dimensional_calibrations=dpc_xdata.dimensional_calibrations[1:])
        except Exception as e:
            import traceback
            print(traceback.format_exc())
            print(e)
            raise

    def commit(self):
        self.computation.set_referenced_xdata("output", self.__result_xdata)


class MakeIDPCMenuItem:
    menu_id = "_processing_menu"
    menu_item_name = _("Make iDPC from DPC")

    def __init__(self, api):
        self.__api = api

    def menu_item_execute(self, window: API.DocumentWindow) -> None:
        document_controller = window._document_controller
        display_item = document_controller.selected_display_item
        data_item = display_item.data_items[0] if display_item and len(display_item.data_items) > 0 else None

        if not data_item:
            return

        api_data_item = Facade.DataItem(data_item)

        if (api_data_item.xdata.is_sequence or api_data_item.xdata.collection_dimension_count == 1) and api_data_item.xdata.datum_dimension_count == 2:
            crop_region = None
            for graphic in api_data_item.graphics:
                if graphic.graphic_type == 'rect-graphic':
                    crop_region = graphic
                    break
            if crop_region is None:
                crop_region = api_data_item.add_rectangle_region(0.5, 0.5, 0.75, 0.75)
            crop_region.label = 'Crop'
            result_data_items = {"output": self.__api.library.create_data_item(title="iDPC of " + data_item.title)}
            self.__api.library.create_computation("nion.make_idpc",
                                                  inputs={"src": api_data_item,
                                                          "gradient_x_index": 0,
                                                          "gradient_y_index": 1,
                                                          "rotation": "None",
                                                          "crop_region": crop_region},
                                                  outputs=result_data_items)
            for data_item in result_data_items.values():
                window.display_data_item(data_item)


class MakeIDPCExtension:

    # required for Swift to recognize this as an extension class.
    extension_id = "nion.extension.make_idpc"

    def __init__(self, api_broker):
        # grab the api object.
        api = api_broker.get_api(version="1", ui_version="1")
        # be sure to keep a reference or it will be closed immediately.
        self.__idpc_menu_item_ref = api.create_menu_item(MakeIDPCMenuItem(api))

    def close(self):
        self.__idpc_menu_item_ref.close()


Symbolic.register_computation_type("nion.make_idpc", MakeIDPC)